package ACL2::Demostracion;

#
# Demostración de la raíz del proyecto, con dependencias cumplidas
# Copyright (C) 2005-2008 Antonio García Domínguez
# $Id$
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use 5.0006;
use warnings;
use strict;
use utf8;

use Carp;

use XML::Writer;    # XML
use Cwd 'realpath'; # Manejador de directorios, rutas y nombres de fichero
use File::ShareDir 'dist_dir'; # Directorio de ficheros compartidos - contiene el DTD
use FindBin;        # Obtiene la ruta del script

use ACL2::Dependencias;
use ACL2::Localizacion;
use ACL2::Tags;
use ACL2::Orden::Fabrica;
use ACL2::Orden;
use ACL2::AnalisisLisp;
use ACL2::Tags;
use XML::Writer;
use File::Spec;

my $lh = ACL2::Localizacion->get_handle();
my $path_to_dtd;

BEGIN {
    # We need to detect the path to the DTD file _before_ any directory
    # changes take place. This is mostly useful during development, but
    # doesn't hurt the rest of the time.

    # Path to the system-wide auto folder with the DTD
    my $dtd_carpeta = eval { dist_dir('ACL2-Procesador') } || '.';
    my $dtd_fichero = "script.dtd";

    # List of all possible paths to the DTD
    my @rutasDTD = ("share/${dtd_fichero}",               # For testing
		    "$FindBin::Bin/share/${dtd_fichero}", # For development/PAR builds
		    "$dtd_carpeta/${dtd_fichero}",        # After install
		    );

    # Use the first existing path
    foreach my $ruta (@rutasDTD) {
	if (-r $ruta) {
	    $path_to_dtd = realpath($ruta);
	    last;
	}
    }
    die $lh->maketext("No se halló el DTD: [_1]", ${path_to_dtd}) 
	if !defined($path_to_dtd);
}

=head1 FUNCTIONS

Abajo se listan todas las funciones de este mE<oacute>dulo.

=head2 new

  Uso: Crea un nuevo Procesador, sin lanzarlo.
  Responsabilidad: Constructor.
  ParE<aacute>metros: Hash con los siguientes elementos:

=over 2

=item *

lisp: Cadena con el contenido del fichero Lisp sin procesar. Si se
omite, se leerá el contenido del fichero bajo la ruta C<ruta_lisp>.

=item *

acl2: Cadena con el contenido de la salida de ACL2 sin procesar. Si se
omite, se leerá el contenido del fichero bajo la ruta C<ruta_dem>.

=item *

ruta_lisp: Cadena con la ruta completa a la fuente Lisp, para posterior referencia.

=item *

ruta_dem: Cadena con la ruta completa a la demostración de ACL2, para posterior referencia.

=item *

deps: Objeto ACL2::Dependencias con las dependencias del fichero actual.

=back

  Devuelve: ACL2::Demostracion construido.
  Excepciones: ninguna.

=cut

sub new {
    my ($clase, %params) = @_;

    # Nos aseguramos de que la ruta al fuente Lisp sea absoluta
    my $d = bless {
	"lisp"      => $params{lisp},
	"acl2"      => $params{acl2},
	"ruta_lisp" => $params{ruta_lisp},
	"ruta_dem"  => $params{ruta_dem},
	"deps" => $params{deps},
    }, $clase;

    $d->_inicializar;
    return $d;
}

=head2 escribir_xml

Uso: vuelca la información obtenida a partir del análisis de la
demostración de ACL2 en formato XML.

Parámetros:

=over 2

=item *

writer: Opcional. Indica el objeto XML::Writer sobre el que se debería
producir la salida. Se creará automáticamente a partir de los otros
parámetros si no se pasa.

=item *

destino_xml: Obligatorio si no se ha pasado C<writer>. Indica a dónde
va a parar el documento XML resultante. Puede ser o bien una instancia
de alguna subclase de IO::Handle, o una referencia a una cadena.

=item *

incrustar_dtd: 1 si se desea incrustar el contenido del DTD en el
propio documento, pudiendo abrir y validarlo en cualquier momento a
costa de un mayor tamaño, o 0 de lo contrario.

=back

Devuelve: nada.

=cut

sub escribir_xml
{
    my ($this, %args) = @_;

    ########## XML::Writer
    my $writer;
    if (exists $args{writer}) {
	$writer = $args{writer};
    } else {
	my $opt_outfile = $args{destino_xml};

	$writer = new XML::Writer(ENCODING => "utf-8", 
				  DATA_MODE => 1, 
				  DATA_INDENT => 2,
				  defined($opt_outfile) ? (OUTPUT => $opt_outfile) : ());
    }

    ########## Declaración XML y DOCTYPE
    $writer->xmlDecl();

    if ($args{incrustar_dtd}) {
	# Generamos DOCTYPE con DTD inline nosotros mismos
	# (XML::Writer sólo imprime públicos y de sistema)
	my $out = $writer->getOutput();
	print $out "<!DOCTYPE script [" . _volcar_fichero($path_to_dtd) . "]>";
    } 
    else {
	# Enlazamos con $dtd_fichero del directorio de ficheros compartidos
	$writer->doctype(ACL2::Tags::MAIN_TAG, undef, $path_to_dtd);
    }

    ########## Cuerpo

    # Tag inicial
    $writer->startTag(ACL2::Tags::MAIN_TAG, 
                      "acl_version"  => $this->{version},
                      "initOpts"     => $this->{init_opts},
		      "source"       => $this->{ruta_lisp},
		      "source_proof" => $this->{ruta_dem});

    # Añadimos la información de dependencias
    $this->{deps}->escribir_xml($writer);

    # Imprimimos la salida de cada una de las S-expresiones
    foreach my $orden (@{$this->{ordenes}}) {
	$orden->escribir_xml($writer);
    }

    ########## Fin
    $writer->endTag(ACL2::Tags::MAIN_TAG);
    $writer->end();
}

##################################################
#
# MÉTODOS PRIVADOS
#
##################################################

sub _inicializar
{
    my ($this, $writer) = @_;

    # Usamos el guión y la salida proporcionada, o si no se dió
    # ninguna, los cargamos desde ficheros
    my $guionLisp  = ($this->{lisp} || _volcar_fichero($this->{ruta_lisp})) 
	or croak $lh->maketext("No se pudo abrir [_1]", $this->{ruta_lisp});
    my $salidaACL2 = ($this->{acl2} || _volcar_fichero($this->{ruta_dem}))
	or croak $lh->maketext("No se pudo abrir [_1]", $this->{ruta_dem});
    $salidaACL2 = __PACKAGE__->_limpiar_salida_ACL2($salidaACL2);
    
    # Separamos la salida de ACL2 según líneas de prompt
    my @salidas = split /(\S*?) \s [pP!s]*>(?=[^>])/x, $salidaACL2;
    
    # Procesamos la presentación
    $salidas[0]  =~ /ACL2 \s Version \s ([\.0-9]+)/xs 
        or die $lh->maketext("Se esperaba la versión de ACL2");
    $this->{version} = $1;
    
    $salidas[0]  =~ /Initialized \s with \s (\(.+?\))/xs
       or die $lh->maketext("Se esperaban las opciones de ACL2");
    $this->{init_opts} = $1;
        
    # Retiramos la bienvenida y despedida
    shift @salidas;
    pop @salidas;

    # Los elementos pares son los paquetes bajo los cuales 
    # están definidas cada una de las órdenes
    my @paquetes = ();
    for (my $i=$#salidas; $i>=0; --$i) {
	unshift(@paquetes, splice(@salidas, $i, 1)) if $i % 2 == 0;
    }
    my @sexprs = ACL2::AnalisisLisp->analizar_fuente_lisp($guionLisp);

    # Procesamos la salida de cada una de las S-expresiones
    $this->{ordenes} = [];
    for (my $i=0; $i <= $#salidas; ++$i) {
	my $orden = ACL2::Orden::Fabrica->crear($sexprs[$i], $salidas[$i], $paquetes[$i]);
	push @{$this->{ordenes}}, $orden;
    }
}

=head1 _limpiar_salida_ACL2

  Uso: ME<eacute>todo estE<aacute>tico privado.
  Responsabilidades: Procesa la salida de ACL2, simplificando su
    procesamiento sin pE<eacute>rdida de informaciE<oacute>n.
  ParE<aacute>metros: salida - cadena con salida a procesar.
  Devuelve: cadena procesada.

=cut

sub _limpiar_salida_ACL2
{
    my ($paquete, $salida) = @_;
    caller eq __PACKAGE__ or die;

    # Buscamos las regiones que hemos de preservar y limpiamos las demás
    my $nuevaSalida = "";
    my $inicio      = 0;
    my $fin         = 0;
    my $finAnterior = 0;
    while ($salida =~ /
	     ^\([^.]+\n       # Línea inicial sin punto (evitamos predicados cortos)
	        [^\.]+?\)\.$  # Resto de líneas hasta el punto final
	   |
	    (?<=\n)\([^\.]+?\)(?=\n\n) # Párrafo aislado y con una larga S-expresión
	   |
	    ^\({2,} (?:.|\n)*? \){2,}\.$   # Inicio de una lista de listas
	   /gmx) 
    {
	$inicio = $-[0];
	$fin    = $+[0];

	my $prefijo      = substr $salida, $finAnterior, $inicio - $finAnterior;
	my $coincidencia = substr $salida, $inicio, $fin - $inicio;

	# Comprobamos ahora si es realmente un pE<aacute>rrafo de s-expresión
	my @listaLisp = ACL2::AnalisisLisp->dividir_lista("(${coincidencia})");
	# Añadimos el punto final al último elemento, si es que lo había
	{
	    my $last = pop @listaLisp;
	    if ($last eq ".") {
		push @listaLisp, (pop @listaLisp) . $last;
	    }
	    else {
		push @listaLisp, $last;
	    }
	}

	if (@listaLisp > 1) {
	    # No es una lista, sino que contiene a una
	    foreach (@listaLisp) {
		/^\(.+\n(?:.|\n)+\)\.$/ && do {
		    # Es la lista que estaba contenida
		    $coincidencia = $_;
		    next;
		};

		$prefijo .= " " . $_;
	    }
	}

	$nuevaSalida .= __PACKAGE__->_limpiar_parrafo($prefijo) . "\n\n";
	$nuevaSalida .= $coincidencia . "\n\n";

	$finAnterior = $fin;
    }
    # Añadimos la cola del documento
    $nuevaSalida .= __PACKAGE__->_limpiar_parrafo(substr $salida, $fin);
    
    return $nuevaSalida;
}

=head1 _limpiar_parrafo

  Uso: ME<eacute>todo estE<aacute>tico privado.
  Responsabilidad: 
    Limpia un trozo de texto puro de ACL2, sin S-expresiones
    indentadas en su interior.
  ParE<aacute>metros: cadena con bloque de texto a limpiar.
  Devuelve: cadena con bloque de texto procesado.

=cut

sub _limpiar_parrafo
{
    my ($paquete, $texto) = @_;
    caller eq __PACKAGE__ or die;

    # Simplificamos las RE: simplificamos todos los espacios entre
    # palabras y retiramos el espaciado inicial y final de cada línea.
    $texto =~ s/(?<!\n)\n(?!\n)/ /g;
    $texto =~ s/[ \t\r]+/ /gm;
    $texto =~ s/(?:^ +| +$)//gm;

    # Arreglamos nombres de regla esparcidos por saltos de línea tras
    # alguno de sus guiones
    $texto =~ s/-\s+(?=\S)/-/g;

    # Retiramos los guiones en el caso de que vengan tras un indicador de paquete
    $texto =~ s/::-/::/g;
    
    # Retiramos barras inversas (con espacio en blanco asociado)
    $texto =~ s/\\\s+//g;

    return $texto;
}

sub _volcar_fichero {
    my ($ruta, $sref) = @_;
    open FICH, "<", $ruta or return undef;
    return join '', <FICH>;
}

1; # End of ACL2::Demostracion
__END__
